diary sesion_6_diario.txt
diary on

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Curso de Introducción al Cálculo Científico con Octave
%
% Guión para la quinta sesión: Valores propios y descomposición de matrices. Resolución de ecuaciones no lineales.
%
% eig, det, inverse, rank, expm, qr, chol, schur, lu,  librería LAPACK, roots, poly, fsolve, 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Comando directo de Octave para el cálculo de valores y vectores propios

A =[ 2, 5, -1 ; 3, 0, 2 ; 0, 5, 3 ];
eig(A)  % la salida son los valores propios

% Para obtener también los vectores propios, hay que "pedirle" las dos salidas:

[P,D] = eig(A)  % P = matriz de vectores propios, D = matriz diagonal con los valores propios

% ojo:  los autovalores no tienen por qué estar ordenados. ¡Véase la ayuda!

P*D-A*P % si A es diagonalizable

A=eye(3);A(1,2)=1;
A
[P,D] = eig(A)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Un ejemplo: deformación de una varilla vertical.

%Al estudiar la deformación de una varilla vertical sometida a fuerzas de 
%compresión verticales (Euler S. XVIII) se obtiene la siguiente ecuación diferencial:

%u'' = c u, con valores en la frontera u(0) = u(L) = 0,

%siendo L la longitud de la varilla y c un valor negativo directamente proporcional a la fuerza que se aplica. 
%La solucion  u(x)=0 corresponde con el caso en que la columna no se deforma; por lo tanto el problema 
%interesante es buscar los valores de c (fuerza aplicada) para los cuales sí hay una deformación u(x) no nula, 
%es decir, un problema de valores propios. Concretamente, el valor propio más interesante es el más pequeño 
%(en módulo), ya que se corresponderá con la mínima fuerza necesaria para que la columna se deforme. 
Si discretizamos el problema usando diferencias finitas centradas, el problema se reduce a calcular los valores 
%propios de cierta matriz (tridiagonal Toeplitz, con -2 en la diagonal, 1 en la superdiagonal y en la subdiagonal).  

L=5; % longitud de la varilla

m=7;  % número de nodos interiores. En tal caso, h=L/(m+1).
% 
v=[-2,1,zeros(1,m-2)];
A=toeplitz(v)

eig(A)

% Además, la función propia correspondiente a dicho valor propio nos da la deformación de la varilla.

% ¿Qué relación hay entre los valores propios del problema discretizado, y los valores propios del problema continuo?
% ¿Qué significan los demás valores propios? 

% Vamos a definir una función, cuya variable será el número de nodos, que calcule los tres primeros valores propios y 
% dibuje las correspondientes funciones propias, es decir, la deformación que sufre la varilla al aplicarle esas fuerzas.

%%% empieza la función 'varilla' %%%%%%%%%%
%%% La guardaremos en un archivo llamado varilla.m  %%%%%%%%%%

function c=varilla(m)

%m es el número de nodos interiores. En tal caso, h=L/(m+1).

L=5; % longitud de la varilla. También podemos ponerla como variable de la función.

h=L./(m+1);
x=linspace(0,L,m+2);
% 
v=[-2,1,zeros(1,m-2)];
A=1./h.^2.*toeplitz(v);

[P,Q]=eig(A);  % Aquí calculamos todos los valores propios, aunque luego sólo vamos a usar tres! 

DQ=diag(Q);
c1=max(DQ);
pos1=find(DQ==c1);

v1=P(:,pos1);
v1=[0,v1',0];
subplot(1,3,1);
plot(v1,x)
str = sprintf("c=%f", c1);
title('Primera autofuncion');
xlabel(str);

DQ(pos1)=min(DQ)-1;
c2=max(DQ);
pos2=find(DQ==c2);

v2=P(:,pos2);
v2=[0,v2',0];
subplot(1,3,2);
plot(v2,x,0,L/2,'*')
str = sprintf("c=%f", c2);
title('Segunda autofuncion');
xlabel(str);

DQ(pos2)=min(DQ)-1;
c3=max(DQ);
pos3=find(DQ==c3);

v3=P(:,pos3);
v3=[0,v3',0];
subplot(1,3,3);
plot(v3,x,[0,0],[L/3,2*L/3],'*')
str = sprintf("c=%f", c3);
title('Tercera autofuncion');
xlabel(str);

c=[c1,c2,c3];

endfunction

%%%%% acaba la función 'varilla' %%%%%%%%
%%%%%%%%%%%%%


% A continuación verificamos como los autovalores se estabilizan en los correspondientes autovalores del problema continuo.

for m=10:10:150 sleep(0.3); varilla(m) endfor

%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Factorización QR de matrices. Tiene aplicación en resolucion de problemas de mínimos cuadrados. También para cálculo de valores propios. 

A =[ 2, 5, -1 ; 3, 0, 2 ; 0, 5, 3 ];

[Q,R] = qr(A) % Q es ortogonal y R triangular superior
Q*R-A

[Q,R,P] = qr(A) % Factorización QR permutada. Los elementos de la diagonal de R están en orden decreciente de magnitud.
Q*R-A*P

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Vamos a programar nosotros un método para calcular los valores y vectores propios, basado en la descomposición QR de la matriz

% Objetivo: generar una sucesión A^(k) de matrices semejantes a A=A^(0), que convergan a una matriz triangular superior. 
% Idea: Factorizamos A^(k)=Q^(k+1) R^(k+1), y la siguiente iteración será A^(k+1)=R^(k+1) Q^(k+1).
% Nótese que A^(k+1)=Q^(k+1) A^(k) Q^(k+1), por tanto son matrice semejantes y tienen los mismos valores propios.
% Criterios de parada: número máximo de iteraciones y un criterio de tolerancia sobre "lo que le sobra" a A^(k) para ser trianglular superior: la norma infinito de la parte inferior de A^(k) sea menor que cierta tolerancia prefijada.
% Este método será convergente siempre que los valores propios sean de módulo diferente.

% Guardaremos la siguiente función en un archivo con el nombre qrvp.m

%%% comienza la función qrvp %%%%
function d = qrvp(A, tol, nmax)

% Esta función calcula una aproximación de los valores propios de una matriz cuadrada A mediante el método QR. 
% nmax determina el número máximo de iteraciones
% tol es una tolerancia prefijada sobre lo que le sobra a las matrices para ser triangulares superiores

T = A;
niter = 0;
test = norm( tril(T,-1) , inf );

while niter <= nmax && test >= tol
      [Q,R] = qr(T);
      T = R*Q;
      niter++;
      test = norm( tril(T,-1) , inf );
      end
if niter > nmax
      printf('El método no ha convergido en %d iteraciones \n', nmax); 
      else
      printf('El método converge en %d iteraciones. \n', niter)
      end
d = diag(T);
end
%%% Acaba la función qrvp %%%%

% Ahora ejecutamos la función qrvp sobre la matriz A
A =[ 2, 5, -1 ; 3, 0, 2 ; 0, 5, 3 ];
eig(A)
qrvp(A,10^-5,100)

% Recordemos como funciona el printf
a = 2;
b = 3;
printf(' vamos a enseñar a %d y b %d \n ', b,a)  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Otras descomposiciones de matrices

% Factorización LU (Si A es full, usa las librerías LAPACK. Si A es sparse, usa las librerías UMFPACK.

A =[ 2, 5, -1 ; 3, 0, 2 ; 0, 5, 3 ];
lu(A) % LU doolitle
[L,U] = lu(A) % L permutación por filas de una triangular inferior
L*U-A
[L,U,P] = lu(A) % L triangular inferior, P es la matriz de permutación
L*U-P*A

% Factorización de Cholesky

M = [5 , 1, -2; 1,7,0; -2,0,9  ];
[R,p]=chol(M)
R' * R - M

M = [5 , 1, -2; 1,7,0; -2,0,-9  ]; % no es definida positiva
R=chol(M)
[R,p]=chol(M)

M = [5 , 3, -2; 1,7,0; -2,0,9  ];  % no es simétrica
R=chol(M)   % ¿¿¿¿????? No detecta la simetría. Asume cierta la simetría y sólo usa la parte triu de M.
R' * R - M

M = [5 , 1, -2; 3,7,0; 2,1,9  ];  
R=chol(M);

% Para jugar un poco, vamos a hacer una nueva función para comprobar la simetría
% Guardamos el siguiente archivo con el nombre ourchol.m

function [y1,y2]=ourchol(A)
M=A;
if M==M' 
[y1,y2]=chol(M);
else
error ( 'Matriz no simetrica');
end
end

% Descomposición de Schur A=U S U' (S triangular superior, U unitaria)

A =[ 2, 5, -1 ; 3, 0, 2 ; 0, 5, 3 ];
[U,S]=schur(A)
U*S*U'-A

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Descomposición en valores singulares A=USV', U y V unitarias

% Si A no es una matriz cuadrada, no podemos calcular los valores propios, pero podemos calcular la descomposición en valores singulares

A =[ 2, 5, -1 ; 3, 0, 2 ; 0, 5, 3 ; 6,0,3 ];
eig(A)
svd(A)
[U,S,V] = svd(A)
U*S*V'-A


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'
%
% Resolución de ecuaciones no lineales
%
% Con Octave podemos resolver ecuaciones de la forma f(x)=0
%
% El comando fsolve (basado en la subrutina hybrd de MINPACK)

function y = f1(x)
y = x^2-2;
end

fsolve("f1",1)
fsolve(@f1,1)
[X, FVEC, INFO, OUTPUT, FJAC] = fsolve("f1",1)

% otro ejemplo

function y = f2(x)
y = log(x)-sin(x);
end
[X, FVEC, INFO, OUTPUT, FJAC] = fsolve("f2",1)

% otra forma de definir la funcion (como funcion handle)

f3=@(x) log(x)-sin(x);
fplot("[log(x),sin(x)]",[1,4])
[X, FVEC, INFO, OUTPUT, FJAC] = fsolve(f3,1)

% Aunque en realidad, para resolver una sola ecuacion con una sola incógnita, es mejor idea usar "fzero" (usa una combinacion de biseccion y secante)

fzero("f1",[1,2])  % le damos un intervalo donde hay una raíz
[X, FVAL, INFO, OUTPUT] = fzero("f1",[1,2]) 
[X, FVAL, INFO, OUTPUT] = fzero("f1",10) % si solo se le da un punto, busca un intervalo cercano donde haya cambio de signo. Si no lo encuentra, da error

% Resolvamos un sistema de ecuaciones no lineales

function y = F1(x)
y = zeros (2, 1);
y(1) = -2*x(1)^2 + 3*x(1)*x(2)   + 4*sin(x(2)) - 6;
y(2) =  3*x(1)^2 - 2*x(1)*x(2)^2 + 3*cos(x(1)) + 4;
end

[X, FVEC, INFO, OUTPUT, FJAC] = fsolve("F1",[1;2])

% Si queréis podemos programar para nuestros alumnos el método de bisección o el de la secante

diary off